@@ -9,7 +9,6 @@ from io import BytesIO
 import hashlib
 import json
 import os
-import tempfile
 
 import requests
 
@@ -98,6 +97,11 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
             self.__base_uri = base_uri
         self.__verify_certs = verify
 
+    def has_api_token(self):
+        """Is there an API Token associated with this client object?
+        """
+        return self.__api_token is not None
+
     def add_client_headers(self, headers=None):
         """Add request headers associated with this client.
 
@@ -206,15 +210,17 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
             raise Error(res.text)
         return payload
 
-    def get_access_rules(self, to_user=None, wdeployment_id=None):
+    def get_access_rules(self, wdeployment_id=None, to_user=None):
         """Get list of access control rules of workspace deployments.
         """
         params = dict()
         if to_user is not None:
             params['to_user'] = to_user
+        if wdeployment_id is None:
+            url = self.__base_uri + '/rules'
         if wdeployment_id is not None:
-            params['wdeployment'] = wdeployment_id
-        res = requests.get(self.__base_uri + '/rules', params=params,
+            url = self.__base_uri + '/deployment/{}/rules'.format(wdeployment_id)
+        res = requests.get(url, params=params,
                            headers=self.__headers, verify=self.__verify_certs)
         if res.ok:
             payload = res.json()
@@ -230,33 +236,29 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
             raise Error(payload)
         return payload['rules']
 
-    def add_access_rule(self, wdeployment_id, capability, to_user=None):
+    def create_access_rule(self, wdeployment_id, capability, to_user=None):
         """Add access control rule for workspace deployment.
 
-        This operation is idempotent, i.e., adding the same rule more
-        than once has no effect.
+        Return rule ID if success.
         """
-        self._modify_access_rule(wdeployment_id=wdeployment_id, capability=capability, to_user=to_user, action='add')
-
-    def del_access_rule(self, wdeployment_id, capability, to_user=None):
-        """Delete access control rule from workspace deployment.
-
-        It is an error to try to delete a rule that does not exist.
-        """
-        self._modify_access_rule(wdeployment_id=wdeployment_id, capability=capability, to_user=to_user, action='del')
-
-    def _modify_access_rule(self, wdeployment_id, capability, action, to_user=None):
         body = {
-            'do': action,
-            'wd': wdeployment_id,
             'cap': capability,
         }
         if to_user is not None:
             body['user'] = to_user
-        res = requests.post(self.__base_uri + '/rule', data=json.dumps(body),
+        res = requests.post(self.__base_uri + '/deployment/{}/rule'.format(wdeployment_id), json=body,
                             headers=self.__headers, verify=self.__verify_certs)
         if not res.ok:
-            raise Error(res.text)
+            raise Error('{} {}'.format(res.status_code, res.reason))
+        return res.json()['rule_id']
+
+    def del_access_rule(self, wdeployment_id, rule_id):
+        """Delete access control rule from workspace deployment.
+        """
+        res = requests.delete(self.__base_uri + '/deployment/{}/rule/{}'.format(wdeployment_id, rule_id),
+                              headers=self.__headers, verify=self.__verify_certs)
+        if not res.ok:
+            raise Error('{} {}'.format(res.status_code, res.reason))
 
     def get_instances(self, include_terminated=False, page=None, max_per_page=None):
         """Get list of your instances.
@@ -327,7 +329,7 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
                 raise Error(errmsg)
             raise Error(payload)
 
-    def request_instance(self, wdeployment_id, sshkey=None, vpn=False, reserve=False, event_url=None, duration=None):
+    def request_instance(self, type_or_wdeployment_id, sshkey=None, vpn=False, reserve=False, event_url=None, duration=None):
         """Request new workspace instance.
 
         If given, sshkey is the public key of the key pair with which
@@ -349,9 +351,9 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
         if event_url is not None:
             body['eurl'] = event_url
         if body:
-            res = requests.post(self.__base_uri + '/new/' + wdeployment_id, data=json.dumps(body), headers=self.__headers, verify=self.__verify_certs)
+            res = requests.post(self.__base_uri + '/new/' + type_or_wdeployment_id, data=json.dumps(body), headers=self.__headers, verify=self.__verify_certs)
         else:
-            res = requests.post(self.__base_uri + '/new/' + wdeployment_id, headers=self.__headers, verify=self.__verify_certs)
+            res = requests.post(self.__base_uri + '/new/' + type_or_wdeployment_id, headers=self.__headers, verify=self.__verify_certs)
         if res.ok:
             payload = res.json()
         else:
@@ -572,7 +574,9 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
             dformat = dformat.lower()
             assert dformat in ['ndarray', 'jpeg']
         res = requests.get(self.__base_uri + '/addon/cam/{}/{}/img'.format(instance_id, camera_id), headers=self.__headers, verify=self.__verify_certs)
-        if not res.ok and res.status_code != 404:
+        if not res.ok:
+            if res.status_code == 404:
+                raise Error('instance {} not found, or cam add-on not active'.format(instance_id))
             raise Error(res.text)
         payload = res.json()
         if not payload['success']:
@@ -666,304 +670,64 @@ class APIClient(object):  # pylint: disable=too-many-public-methods
         if not res.ok:
             raise Error(res.text)
 
-
-class Instance(object):  # pylint: disable=too-many-public-methods,too-many-instance-attributes
-    """Manager for a workspace instance
-    """
-    def __init__(self, workspace_types=None, wdeployment_id=None, instance_id=None, api_token=None, headers=None, apic=None):
-        """client for a workspace instance
-
-        At least one of workspace_types or wdeployment_id must be
-        given. If both are provided (not None), consistency is
-        checked: the type of the workspace deployment of the given
-        identifier is compared with the given type. If they differ, no
-        instance is created, and ValueError is raised.
-
-        If instance_id is given, then attempt to attach this class
-        to an existing instance. In this case, neither workspace_types
-        or wdeployment_id is required. If they are provided, then
-        consistency is checked.
-
-        The optional parameter `apic` is an instance of APIClient. If
-        it is not given, then an APIClient object is instantiated
-        internally from the parameters `api_token` etc., corresponding
-        to parameters APIClient of the same name.
+    def get_ci_projects(self):
+        """Get list of your CI projects.
         """
-        # pylint: disable=too-many-branches,too-many-arguments
-        if workspace_types is None and wdeployment_id is None and instance_id is None:
-            raise ValueError('at least workspace_types, wdeployment_id, or instance_id must be given')
-
-        if apic is None:
-            self.apic = APIClient(api_token=api_token, headers=headers)
+        res = requests.get(self.__base_uri + '/ci/projects', headers=self.__headers, verify=self.__verify_certs)
+        if res.ok:
+            payload = res.json()
         else:
-            self.apic = apic
+            try:
+                payload = res.json()
+            except:
+                raise Error(res.text)
+            if 'error_message' in payload:
+                if payload['error_message'] == 'wrong authorization token':
+                    raise WrongAuthToken('wrong authorization token')
+                raise Error(payload['error_message'])
+            raise Error(payload)
+        return payload
 
-        if wdeployment_id is not None and workspace_types is not None:
-            x = self.apic.get_wdeployment_info(wdeployment_id)
-            if x['type'] not in workspace_types:
-                raise ValueError('workspace deployment {} does not have type in {}'.format(wdeployment_id, workspace_types))
-
-        if workspace_types is not None:
-            candidates = self.apic.get_wdeployments(types=workspace_types)
-            if not candidates:
-                raise ValueError('no deployments found with any type in {}'.format(workspace_types))
-            self._wdeployment_id = candidates[0]['id']
+    def create_ci_project(self, repo_url):
+        """Create CI project.
 
+        requires a repository URL.
+        """
+        res = requests.post(self.__base_uri + '/ci/new', json={'repo_url': repo_url}, headers=self.__headers, verify=self.__verify_certs)
+        if res.ok:
+            payload = res.json()
         else:
-            self._wdeployment_id = wdeployment_id
+            try:
+                payload = res.json()
+            except:
+                raise Error(res.text)
+            if 'error_message' in payload:
+                if payload['error_message'] == 'wrong authorization token':
+                    raise WrongAuthToken('wrong authorization token')
+                raise Error(payload['error_message'])
+            raise Error(payload)
+        return payload['pid']
 
-        self._type = None
-
-        if instance_id is None:
-            payload = self.apic.request_instance(self._wdeployment_id, reserve=False)
-            self._id = payload['id']
-            self._status = 'INIT'  # Instance always begins at INIT
-            if 'sshkey' in payload:
-                self.__sshkey = payload['sshkey']
-            else:
-                self.__sshkey = None
+    def submit_ci_job(self, pid, branch, ref):
+        """Create CI project.
 
+        requires a repository URL.
+        """
+        res = requests.post(self.__base_uri + '/ci/project/{}/job'.format(pid),
+                            json={'branch': branch, 'commit': ref},
+                            headers=self.__headers, verify=self.__verify_certs)
+        if res.ok:
+            payload = res.json()
         else:
-            self._id = instance_id
-            payload = self.apic.get_instance_info(self._id)
-            if self._wdeployment_id is not None:
-                assert payload['deployment'] == self._wdeployment_id
-            else:
-                self._wdeployment_id = payload['deployment']
-            self._status = payload['status']
-            self.__sshkey = None
-
-        self._details = None
-
-        self._conn = None
-        self.__sshclient = None
-        self.__sftpclient = None
-
-
-    def get_wdeployment_info(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.get_wdeployment_info(self._wdeployment_id)
-
-
-    def get_access_rules(self, to_user=None):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.get_access_rules(to_user=to_user, wdeployment_id=self._wdeployment_id)
-
-    def add_access_rule(self, capability, to_user=None):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.add_access_rule(wdeployment_id=self._wdeployment_id, capability=capability, to_user=to_user)
-
-    def del_access_rule(self, capability, to_user=None):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.del_access_rule(wdeployment_id=self._wdeployment_id, capability=capability, to_user=to_user)
-
-
-    def get_firewall_rules(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.get_firewall_rules(self._id)
-
-    def add_firewall_rule(self, action, source_address=None):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.add_firewall_rule(self._id, action=action, source_address=source_address)
-
-    def flush_firewall_rules(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.flush_firewall_rules(self._id)
-
-
-    def get_vpn_newclient(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.get_vpn_newclient(self._id)
-
-
-    def activate_addon_cam(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.activate_addon_cam(self._id)
-
-    def status_addon_cam(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.status_addon_cam(self._id)
-
-    def get_snapshot_cam(self, camera_id=0, coding=None, dformat=None):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.get_snapshot_cam(self._id, camera_id=camera_id, coding=coding, dformat=dformat)
-
-    def deactivate_addon_cam(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.deactivate_addon_cam(self._id)
-
-
-    def activate_addon_drive(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.activate_addon_drive(self._id)
-
-    def status_addon_drive(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.status_addon_drive(self._id)
-
-    def send_drive_command(self, command):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.send_drive_command(instance_id=self._id, command=command)
-
-    def deactivate_addon_drive(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.deactivate_addon_drive(self._id)
-
-
-    def activate_addon_mistyproxy(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.activate_addon_mistyproxy(self._id)
-
-    def status_addon_mistyproxy(self):
-        """This is a wrapper for APIClient method of same name."""
-        return self.apic.status_addon_mistyproxy(self._id)
-
-    def deactivate_addon_mistyproxy(self):
-        """This is a wrapper for APIClient method of same name."""
-        self.apic.deactivate_addon_mistyproxy(self._id)
-
-
-    def get_status(self):
-        """Get status of this workspace instance.
-
-        For example, the status is `READY` when the instance is ready
-        to be used, and `INIT` when it is initializing.
-        """
-        payload = self.apic.get_instance_info(self._id)
-        if self._wdeployment_id is None:
-            self._wdeployment_id = payload['deployment']
-        if self._details is None:
-            self._details = {
-                'type': payload['type'],
-                'region': payload['region'],
-                'starttime': payload['starttime'],
-            }
-        self._status = payload['status']
-        if 'fwd' in payload:
-            self._conn = {
-                'type': 'sshtun',
-            }
-            if 'ipv4' in payload['fwd']:
-                self._conn['ipv4'] = payload['fwd']['ipv4']
-            if 'port' in payload['fwd']:
-                self._conn['port'] = payload['fwd']['port']
-            if 'hostkeys' in payload:
-                self._conn['hostkeys'] = payload['hostkeys']
-        return self._status
-
-
-    def get_details(self):
-        """Get status and details about this workspace instance.
-        """
-        status = self.get_status()
-        res = self._details.copy()
-        res['status'] = status
-        if self._conn is not None:
-            res['conn'] = self._conn
-        return res
-
-
-    def terminate(self):
-        """Terminate this instance.
-        """
-        self.stop_sshclient()
-        self.apic.terminate_instance(self._id)
-
-
-    def stop_sshclient(self):
-        """Stop, close SSH client connection to instance, if it exists.
-        """
-        if self.__sshclient is not None:
-            self.__sshclient.close()
-            self.__sshclient = None
-            self.__sftpclient = None
-
-
-    def start_sshclient(self):
-        """Create SSH client to instance.
-
-        This method is a prerequisite to exec_ssh(), which executes
-        remote terminal commands.
-        """
-        # pylint: disable=import-outside-toplevel
-        import paramiko
-        status = self.get_status()
-        if status != 'READY':
-            raise Exception('instance not ready')
-        if 'ipv4' not in self._conn or 'port' not in self._conn or 'hostkeys' not in self._conn:
-            self.get_details()
-        host = self._conn['ipv4']
-        port = self._conn['port']
-        hostkey = self._conn['hostkeys'][0]
-
-        fd, keypath = tempfile.mkstemp()
-        fp = os.fdopen(fd, 'wt')
-        fp.write(self.__sshkey)
-        fp.close()
-
-        fd, known_hosts = tempfile.mkstemp()
-        fp = os.fdopen(fd, 'wt')
-        sshhost = '[{IPADDR}]:{PORT}'.format(IPADDR=host, PORT=port)
-        fp.write(sshhost + ' ' + hostkey)
-        fp.close()
-
-        self.__sshclient = paramiko.client.SSHClient()
-        self.__sshclient.load_system_host_keys(known_hosts)
-        pkey = paramiko.rsakey.RSAKey.from_private_key_file(keypath)
-        self.__sshclient.connect(host, port=port, username='root', pkey=pkey, timeout=5)
-
-        os.unlink(keypath)
-        os.unlink(known_hosts)
-
-
-    def exec_ssh(self, command, timeout=None, get_files=False):
-        """Execute command via SSH.
-
-        https://docs.paramiko.org/en/2.4/api/client.html#paramiko.client.SSHClient.exec_command
-
-        If get_files=True, then return files of stdin, stdout, and
-        stderr.
-        """
-        assert self.__sshclient is not None
-        stdin, stdout, stderr = self.__sshclient.exec_command(command, timeout=timeout)
-        if get_files:
-            return stdin, stdout, stderr
-        return stdout.read()
-
-
-    def put_file(self, localpath, remotepath):
-        """Put local file onto remote host.
-
-        For the general case, the underlying Paramiko SFTP object is
-        available from sftp_client().
-        """
-        assert self.__sshclient is not None
-        if self.__sftpclient is None:
-            self.__sftpclient = self.__sshclient.open_sftp()
-        return self.__sftpclient.put(localpath, remotepath)
-
-
-    def get_file(self, remotepath, localpath):
-        """Get file from remote host.
-
-        For the general case, the underlying Paramiko SFTP object is
-        available from sftp_client().
-        """
-        assert self.__sshclient is not None
-        if self.__sftpclient is None:
-            self.__sftpclient = self.__sshclient.open_sftp()
-        return self.__sftpclient.get(remotepath, localpath)
-
-
-    def sftp_client(self):
-        """Get Paramiko SFTP client.
-
-        Note that methods put_file() and get_file() are small wrappers
-        to put() and get() of this Paramiko class.
-
-        Read about it at https://docs.paramiko.org/en/2.4/api/sftp.html
-        """
-        assert self.__sshclient is not None
-        if self.__sftpclient is None:
-            self.__sftpclient = self.__sshclient.open_sftp()
-        return self.__sftpclient
+            if res.status_code == 404:
+                raise Error('not found')
+            try:
+                payload = res.json()
+            except:
+                raise Error(res.text)
+            if 'error_message' in payload:
+                if payload['error_message'] == 'wrong authorization token':
+                    raise WrongAuthToken('wrong authorization token')
+                raise Error(payload['error_message'])
+            raise Error(payload)
+        return payload['jid']
