@@ -26,6 +26,8 @@ try:
 except ImportError:
     from time import time as monotonic_unless_py2
 
+import requests
+
 from . import api as rerobots_api
 from .__init__ import __version__
 
@@ -148,6 +150,9 @@ def add_cli_search(subparsers):
     search_parser.add_argument('-h', '--help', dest='print_search_help',
                                action='store_true', default=False,
                                help='print this help message and exit')
+    search_parser.add_argument('--include-user-provided', dest='include_user_provided',
+                               action='store_true', default=False,
+                               help='include user_provided workspace deployments in search')
     search_parser.add_argument('QUERY', nargs='?', default=None)
 
 
@@ -173,7 +178,7 @@ def add_cli_launch(subparsers):
     launch_parser.add_argument('-h', '--help', dest='print_launch_help',
                                action='store_true', default=False,
                                help='print this help message and exit')
-    launch_parser.add_argument('ID', nargs='?', default=None, help='deployment ID')
+    launch_parser.add_argument('ID', nargs='?', default=None, help='workspace type or deployment ID')
     launch_parser.add_argument('--secret-key', metavar='FILE', dest='secretkeypath',
                                default=None,
                                help=('name of file in which to write new secret key '
@@ -206,6 +211,45 @@ def add_cli_terminate(subparsers):
     terminate_parser.add_argument('ID', nargs='?', default=None, help='instance ID')
 
 
+def add_cli_list_ci_projects(subparsers):
+    """Create subparser for `list-ci-projects`.
+    """
+    desc = 'list all CI projects of this user.'
+    ci_list_proj_parser = subparsers.add_parser('list-ci-projects', description=desc, help=desc, add_help=False)
+    ci_list_proj_parser.add_argument('-h', '--help', dest='print_list_ci_projects_help',
+                                     action='store_true', default=False,
+                                     help='print this help message and exit')
+
+
+def add_cli_create_ci_project(subparsers):
+    """Create subparser for `create-ci-project`.
+    """
+    desc = 'create a CI project for this user.'
+    ci_create_proj_parser = subparsers.add_parser('create-ci-project', description=desc, help=desc, add_help=False)
+    ci_create_proj_parser.add_argument('--repo-url', metavar='URL', dest='repo_url',
+                                       default=None,
+                                       help='URL of the Git repo for CI/CD')
+    ci_create_proj_parser.add_argument('-h', '--help', dest='print_create_ci_project_help',
+                                       action='store_true', default=False,
+                                       help='print this help message and exit')
+
+def add_cli_submit_ci_job(subparsers):
+    """Create subparser for `submit-ci-job`.
+    """
+    desc = 'create a CI project for this user.'
+    ci_submit_job_parser = subparsers.add_parser('submit-ci-job', description=desc, help=desc, add_help=False)
+    ci_submit_job_parser.add_argument('--repo-branch', metavar='BRANCH', dest='repo_branch',
+                                      default=None,
+                                      help='branch of the repo for CI/CD')
+    ci_submit_job_parser.add_argument('--repo-ref', metavar='REF', dest='repo_ref',
+                                      default=None,
+                                      help='commit hash of the Git repo for CI/CD')
+    ci_submit_job_parser.add_argument('ID', default=None, help='project ID')
+    ci_submit_job_parser.add_argument('-h', '--help', dest='print_submit_ci_job_help',
+                                      action='store_true', default=False,
+                                      help='print this help message and exit')
+
+
 def py2_replace_help(argv):
     """
 
@@ -285,6 +329,9 @@ def create_parser(argv=None):
     add_cli_wdinfo(subparsers)
     add_cli_launch(subparsers)
     add_cli_terminate(subparsers)
+    add_cli_list_ci_projects(subparsers)
+    add_cli_create_ci_project(subparsers)
+    add_cli_submit_ci_job(subparsers)
 
     subparsers.add_parser('version', help='print version number and exit.')
     help_parser = subparsers.add_parser('help', help='print this help message and exit')
@@ -486,7 +533,11 @@ def cli_isready(apic, args):
 def cli_search(apic, args):
     """Implement CLI command `search`.
     """
-    for wdinfo in apic.get_wdeployments(query=args.QUERY):
+    if args.include_user_provided:
+        query = apic.get_wdeployments(query=args.QUERY)
+    else:
+        query = apic.get_wdeployments(query=args.QUERY, types=['!user_provided'])
+    for wdinfo in query:
         print('{}\t{}'.format(wdinfo['id'], wdinfo['type']))
     return 0
 
@@ -494,7 +545,10 @@ def cli_search(apic, args):
 def cli_wdinfo(apic, args):
     """Implement CLI command `wdinfo`.
     """
-    print(json.dumps(apic.get_wdeployment_info(args.ID), indent=2))
+    wdinfo = apic.get_wdeployment_info(args.ID)
+    if apic.has_api_token():
+        wdinfo['cap'] = apic.get_access_rules(args.ID)
+    print(json.dumps(wdinfo, indent=2))
     return 0
 
 
@@ -528,6 +582,43 @@ def cli_terminate(apic, args):
     return 0
 
 
+def cli_list_ci_projects(apic, args):
+    """Implement CLI command `list-ci-projects`.
+    """
+    # pylint: disable=unused-argument
+    ci_projs = apic.get_ci_projects()
+    if ci_projs:
+        for pid, attr in ci_projs.items():
+            print('{}:'.format(pid))
+            print('  url: https://ci.rerobots.net/p/{}'.format(pid))
+            print('  repo_url: {}'.format(attr['repo_url']))
+    return 0
+
+
+def cli_create_ci_project(apic, args):
+    """Implement CLI command `create-ci-project`.
+    """
+    # pylint: disable=unused-argument
+    if args.repo_url is None:
+        print('error: missing required argument: --repo-url')
+        return 1
+    proj_id = apic.create_ci_project(args.repo_url)
+    print(proj_id)
+    return 0
+
+
+def cli_submit_ci_job(apic, args):
+    """Implement CLI command `submit-ci-project`.
+    """
+    # pylint: disable=unused-argument
+    if args.repo_branch is None or args.repo_ref is None:
+        print('error: missing required arguments: --repo-branch or --repo-ref')
+        return 1
+    job_id = apic.submit_ci_job(args.ID, branch=args.repo_branch, ref=args.repo_ref)
+    print(job_id)
+    return 0
+
+
 def main(argv=None):
     """Process command-line arguments.
     """
@@ -555,7 +646,7 @@ def main(argv=None):
     else:
         apic = rerobots_api.APIClient()
 
-    if args.command not in ['search', 'wdinfo', 'list', 'info', 'isready', 'addon-cam', 'addon-mistyproxy', 'addon-drive', 'terminate', 'launch']:
+    if args.command not in ['search', 'wdinfo', 'list', 'info', 'isready', 'addon-cam', 'addon-mistyproxy', 'addon-drive', 'terminate', 'launch', 'list-ci-projects', 'create-ci-project', 'submit-ci-job']:
         print('Unrecognized command. Try `help`.')
         return 1
 
@@ -563,7 +654,14 @@ def main(argv=None):
     if getattr(args, 'print_{}_help'.format(command)):
         subparsers.choices[args.command].print_help()
         return 0
-    return globals()['cli_' + command](apic, args)
+    try:
+        return globals()['cli_' + command](apic, args)
+    except rerobots_api.Error as err:
+        print('error: {}'.format(err))
+        return 1
+    except requests.exceptions.ConnectionError:
+        print('api.rerobots.net server cannot be reached.  Are you connected to the Internet?')
+        return 1
 
 
 if __name__ == '__main__':
